@@ -1,18 +1,19 @@
 import datetime
 import importlib
-import inspect
 import re
 import socket
 import time
 
-from irc.buffer import DecodingLineBuffer
+from pyromancer import utils
 
 
 class Pyromancer(object):
 
     def __init__(self, settings_path):
         self.settings = Settings(settings_path)
+        self.setup_database()
         self.find_commands()
+        self.find_timers()
 
     def run(self):
         self.connect()
@@ -22,9 +23,12 @@ class Pyromancer(object):
         self.connection = Connection(self.settings.host, self.settings.port,
                                      self.settings.encoding)
 
+        self.online = True
+        self.connect_time = datetime.datetime.now()
         self.connection.write('NICK {}\n'.format(self.settings.nick))
         self.connection.write('USER {0} {1} {1} :{2}\n'.format(
             self.settings.nick, self.settings.host, self.settings.real_name))
+        self.connection.me.nick = self.settings.nick
 
     def listen(self):
         self.online = True
@@ -36,49 +40,56 @@ class Pyromancer(object):
             for line in self.connection.buffer.lines():
                 self.process(line)
 
+            for timer in self.timers:
+                timer.match(self.timers, self.connect_time, self.connection,
+                            self.settings)
+
             time.sleep(1.0 / ticks)
 
     def process(self, line):
-        line = Line(line)
+        line = Line(line, self.connection)
 
         if line[0] == 'PING':
             self.connection.write('PONG {}\n'.format(line[1]))
 
         for c in self.commands:
-            c.command.match(line, self.connection, self.settings)
+            c.command.match(line, self.timers, self.connection, self.settings)
 
     def find_commands(self):
         self.commands = []
 
-        for package in self.settings.packages:
-            if isinstance(package, tuple):
-                package_settings = package[1]
-                package = package[0]
-            else:
-                package_settings = {}
+        utils.find_functions(
+            self.settings.packages, self.commands, 'commands',
+            'disabled_commands', when=lambda f: hasattr(f, 'command'))
 
-            ignored = package_settings.get('disabled_commands', [])
-            ignored = ['{}.{}'.format(package, i) if not i.startswith(package)
-                       else i for i in ignored]
+    def find_timers(self):
+        self.timers = []
 
-            module_name = '{}.commands'.format(package)
-            if module_name in ignored:
-                continue
+        utils.find_functions(
+            self.settings.packages, self.timers, 'timers', 'disabled_timers',
+            when=lambda f: hasattr(f, 'timer'), ret=lambda f: f.timer)
 
-            module = importlib.import_module(module_name)
+    def setup_database(self):
+        if self.settings.database:
+            from sqlalchemy import create_engine
 
-            modules = [('', module,)]
-            modules.extend(inspect.getmembers(module, inspect.ismodule))
+            from pyromancer.database import Session, Base
 
-            for name, module in modules:
-                if module.__name__ in ignored:
+            engine = create_engine(self.settings.database)
+            Session.configure(bind=engine)
+
+            for package in self.settings.packages:
+                if isinstance(package, tuple):
+                    package = package[0]
+
+                module_name = '{}.models'.format(package)
+
+                try:
+                    importlib.import_module(module_name)
+                except ImportError:
                     continue
 
-                functions = inspect.getmembers(module, inspect.isfunction)
-                self.commands.extend(
-                    f for fn, f in functions if
-                    hasattr(f, 'command') and '{}.{}'.format(
-                        module.__name__, fn) not in ignored)
+            Base.metadata.create_all(bind=engine)
 
 
 class Settings(object):
@@ -123,40 +134,144 @@ class Settings(object):
                              'installed packages'.format(item))
 
 
+class LineBuffer(object):
+    """Line buffer based on irc library's DecodingLineBuffer.
+
+    See https://bitbucket.org/jaraco/irc
+    """
+    line_sep_exp = re.compile(b'\r?\n')
+
+    def __init__(self, encoding='utf-8'):
+        self.buffer = b''
+        self.encoding = encoding
+
+    def feed(self, bytes):
+        self.buffer += bytes
+
+    def lines(self):
+        lines = self.line_sep_exp.split(self.buffer)
+        # save the last, unfinished, possibly empty line
+        self.buffer = lines.pop()
+
+        for line in lines:
+            yield line.decode(self.encoding, 'strict')
+
+    def __iter__(self):
+        return self.lines()
+
+    def __len__(self):
+        return len(self.buffer)
+
+
 class Connection(object):
 
     def __init__(self, host, port, encoding='utf8'):
-        self.socket = socket.socket()
+        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         self.socket.connect((host, port))
+        self.socket.setblocking(False)
         self.encoding = encoding
 
-        self.buffer = DecodingLineBuffer()
-        self.buffer.encoding = self.encoding
+        self.buffer = LineBuffer(self.encoding)
+
+        self.me = User('')
+
+    @property
+    def users(self):
+        users = [self.me]
+        for chan in self.me.channels:
+            users.extend(chan.users)
+
+        return list(set(users))
+
+    @property
+    def channels(self):
+        return self.me.channels
 
     def write(self, data):
         self.socket.send('{}\n'.format(data).encode(self.encoding))
 
     def read(self, bytes=4096):
-        self.buffer.feed(self.socket.recv(bytes))
+        try:
+            self.buffer.feed(self.socket.recv(bytes))
+        except BlockingIOError:
+            pass
 
     def msg(self, target, msg):
         self.write('PRIVMSG {} :{}'.format(target,  msg))
 
 
 class User(object):
-    nick = None
     name = None
+    host = None
 
     def __init__(self, str):
         if '@' in str:
-            self.nick, parts = str.split('!', 1)
-            self.name, self.host = parts.split('@', 1)
+            self.nick, self.name, self.host = self.split_user_str(str)
         else:
-            self.host = str
+            self.nick = str
+
+        self.channels = []
 
     def __repr__(self):
         return '{0.nick}@{0.host}'.format(self) if self.nick else self.host
 
+    @staticmethod
+    def split_user_str(str):
+        nick, parts = str.split('!', 1)
+        name, host = parts.split('@', 1)
+        return nick.lstrip(':'), name, host
+
+    @classmethod
+    def get(cls, str, pool):
+        if isinstance(pool, Match):
+            pool = pool.connection
+
+        if isinstance(pool, Connection):
+            pool = pool.users
+
+        if pool is None:
+            pool = []
+
+        if '@' in str:
+            nick, _, _ = cls.split_user_str(str)
+        else:
+            nick = str
+
+        for user in pool:
+            if user.nick == nick:
+                return user
+
+        return cls(str)
+
+
+class Channel(object):
+
+    def __init__(self, name):
+        self.name = name.lstrip(':')
+        self.users = []
+
+    def __repr__(self):
+        return self.name
+
+    @classmethod
+    def get(cls, name, pool):
+        if isinstance(pool, Match):
+            pool = pool.connection
+
+        if isinstance(pool, Connection):
+            pool = pool.channels
+
+        if pool is None:
+            pool = []
+
+        name = name.lstrip(':')
+
+        for chan in pool:
+            if chan.name == name:
+                return chan
+
+        return cls(name)
+
 
 class Match(object):
     """
@@ -170,10 +285,11 @@ class Match(object):
     things a command may like to do.
     """
 
-    def __init__(self, match, line, connection):
+    def __init__(self, match, line, connection, settings=None):
         self.match = match
         self.line = line
         self.connection = connection
+        self.settings = settings
 
     def __getitem__(self, item):
         try:
@@ -207,7 +323,7 @@ class Match(object):
             kw = {
                 'm': self,
                 'b': chr(2),
-                'k': chr(3),
+                'c': chr(3),
                 'u': chr(31),
             }
             kw.update(kwargs)
@@ -224,13 +340,15 @@ class Match(object):
         self.connection.msg(target, message)
 
 CODE_PATTERN = re.compile(r'\d{3}')
+COMMAND_PATTERN = re.compile(r'[A-Z]{4,5}')
 
 
 class Line(object):
 
-    def __init__(self, data):
-        self.raw = data
+    def __init__(self, data, connection=None):
+        self.raw = data.lstrip(':')
         self.datetime = datetime.datetime.now()
+        self.connection = connection
         self.parse()
 
     def __getitem__(self, item):
@@ -254,9 +372,83 @@ class Line(object):
 
         if self.usermsg:
             # ":Nick!Name@Host PRIVMSG #Chan :Hello world!"
-            self.sender = User(self.parts[0][1:])
             self.target = self.parts[2]
             self.pm = self.target[0] != '#'
             self.full_msg = ' '.join(self.parts[3:])[1:]
         elif CODE_PATTERN.match(self.parts[1]):
             self.code = int(self.parts[1])
+        elif COMMAND_PATTERN.match(self.parts[1]):
+            self.command = self.parts[1]
+
+        if self.usermsg or getattr(self, 'command', None):
+            self.sender = User.get(self.parts[0], self.connection)
+
+            self.channel = None
+            if not getattr(self, 'pm', False):
+                self.channel = Channel.get(self.parts[2], self.connection)
+
+
+class Timer(object):
+
+    def __init__(self, scheduled, msg_or_command=None, *args, **kwargs):
+        self.scheduled = scheduled
+        self.direct = kwargs.pop('direct', False)
+        self.remaining = kwargs.pop('count', 0)
+        target = kwargs.pop('target', None)
+
+        self.msg_tuple = None
+        self.function = None
+
+        if callable(msg_or_command):
+            self.function = msg_or_command
+        else:
+            if target is not None and msg_or_command is not None:
+                self.msg_tuple = (target, msg_or_command, args, kwargs,)
+
+    def __eq__(self, other):
+        return (self.scheduled == other.scheduled and
+                self.function is other.function and
+                self.msg_tuple == other.msg_tuple)
+
+    def match(self, timers, connect_time, connection, settings):
+        if self.matches(connect_time):
+            self.last_time = datetime.datetime.now()
+            match = Match(None, None, connection, settings)
+
+            if self.function is not None:
+                result = self.function(match)
+
+                if result is not None:
+                    self.send_messages(result, match, timers)
+
+            if self.msg_tuple is not None:
+                self.send_messages(self.msg_tuple, match, timers)
+
+            if self.remaining > 0:
+                self.remaining -= 1
+
+                if self.remaining == 0:
+                    timers.remove(self)
+
+    def matches(self, connect_time):
+        if self.direct:
+            self.direct = False
+            return True
+
+        if isinstance(self.scheduled, datetime.datetime):
+            next_time = self.scheduled
+
+        if isinstance(self.scheduled, datetime.timedelta):
+            if hasattr(self, 'last_time'):
+                next_time = self.last_time + self.scheduled
+            else:
+                next_time = connect_time + self.scheduled
+
+        return datetime.datetime.now() >= next_time
+
+    def send_messages(self, result, match, timers):
+        for r in utils.process_messages(result, with_target=True):
+            if isinstance(r, Timer):
+                timers.append(r)
+            else:
+                match.msg(r[1], *r[2], target=r[0], **r[3])
